from .util import PdfFixture


PDF_FIXTURES = [
    PdfFixture(
        name="1.pdf",
        page_count=6,
        file_hash="121ffed4336e6129e97ee3c4cb747864",
        ocr_name="1_ocr.json",
    ),
]
